#ifndef YUV_RGB_COMMON_H_
#define YUV_RGB_COMMON_H_
// Copyright 2016 Adrien Descamps
// Distributed under BSD 3-Clause License

typedef enum
{
    YCBCR_601_FULL,
    YCBCR_601_LIMITED,
    YCBCR_709_FULL,
    YCBCR_709_LIMITED,
    YCBCR_2020_NCL_FULL,
} YCbCrType;

#endif /* YUV_RGB_COMMON_H_ */
